// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of scatter
//
// scatter is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// scatter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the
//       Free Software Foundation, Inc.
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <algorithm>

#include <saotrace/io/fits/fits.h>

namespace FITS = SAOTrace::IO::FITS;

#include "utils.h"

#include "ZoneMap.h"
#include "zone_info.h"


namespace Scatter {

    ZoneMap::Zone::Zone( const ZoneInfo& zi ) :
	zmin( zi.zmin + zi.z0 ),
	zmax( zi.zmax + zi.z0 ),
	id(zi.zoneid),
	in_plane(zi.in_plane),
	out_of_plane(zi.out_of_plane ),
	scale_out_of_plane_by_sin_alpha( false ),
	make_out_of_plane_symmetric( false ),
	scale_in_plane(zi.scale_in_plane),
	scale_out_of_plane(zi.scale_out_of_plane )
    {

	if ( in_plane ) {
	    db_in_plane = ScatterDBPtr(ScatterDB::load(zi.scatdb_in_plane));

	    if ( ! db_in_plane )
		throw( Exception( "in plane scattering requested, but there's no scattering table for it\n" ) );
	}

	if ( out_of_plane ) {

	    db_out_of_plane = ScatterDBPtr( ScatterDB::load(zi.scatdb_out_of_plane) );

	    // if only a single scattering table is available
	    // it is used for both in plane and out of plane.
	    // However, we need to scale it be sin(alpha).
	    // And, if it's not a symmetric distribution, we
	    // need to make it so.

	    if ( ! db_out_of_plane ) {

		db_out_of_plane = db_in_plane
		    ? db_in_plane
		    : ScatterDBPtr(ScatterDB::load(zi.scatdb_in_plane));

		if ( ! db_out_of_plane )
		    throw( Exception( "out of plane scattering requested, but there's no scattering table for it\n" ) );

		scale_out_of_plane_by_sin_alpha = true;

		if ( ! db_out_of_plane->is_symmetric() )
		    make_out_of_plane_symmetric = true;
	    }

	}
    }


    ZoneMap::ZoneMap (
		      const string& db_filename_in,
		      const int surfno,
		      const int shell,
		      const double z0,
		      const bool in_plane,
		      const bool out_of_plane
		      ) throw ( Exception ) {


	string db_filename = string_interp( db_filename_in);

	RDB db( db_filename );

	// map RDB column values to local storage for a single row
	ZoneInfo zi( db, db_filename, z0, in_plane, out_of_plane );

	stringstream mirror_ss;
	mirror_ss << (1 == surfno ? 'p' : 'h' ) << shell;
	string requested_mirror = mirror_ss.str( );

	while ( db.read( ) ) {

	    if ( requested_mirror == zi.mirror ) {

		// interpolate environment variables in just read string values
		zi.interp_env();

		Zone* zp = new Zone( zi );
		_zones.push_back( zp );
	    }
	}


	if ( ! _zones.size( ) )
	    throw( Exception( string("mirror '") + requested_mirror + "' not found in scatter database:  " + db_filename ) );

	// now, sort the zones to ensure that they're in ascending order...
	sort( _zones.begin( ), _zones.end( ), Zone::cmp() );

	//  check that zones don't overlap...

	for ( size_t i = 0 ; i < _zones.size() - 1 ; i++ ) {

	    if ( _zones[i]->zmin > _zones[i+1]->zmin || _zones[i]->zmax > _zones[i+1]->zmax ) {

		ostringstream ss;
		ss << "zones " << _zones[i]->id << " and " << _zones[i+1]->id << " overlap";
		throw( Exception( ss.str( ) ) );

	    }
	}

    }

    ZoneMap::~ZoneMap( ) {

	Zones::iterator zone = _zones.begin();
	Zones::iterator end  = _zones.end();

	for ( ; zone != end ; ++zone )
	    delete *(zone);

    }


    ZoneMap::Zone* ZoneMap::findZone( double z ) {


	Zones::iterator begin = _zones.begin();
	Zones::iterator end   = _zones.end();

	Zones::iterator zone = lower_bound( begin, end, z, Zone::cmp() );

	if ( zone == end
	     || ( z < (*zone)->zmin && zone != begin )
	    ) {
	    --zone;
	}

	return ( (*zone)->zmin > z || (*zone)->zmax < z )  ? NULL : (*zone);
    }

}
